from pathlib import Path
from typing import List, Optional

from .accounts import Account
from .client import SafewayClient
from .email import email_clip_results, email_error
from .errors import ClipError, Error, TooManyClipErrors
from .models import Offer, OfferStatus
from .utils import yield_delay

CLIP_ERROR_MAX = 5


class SafewayCoupons:
    def __init__(
        self,
        send_email: bool = True,
        sendmail: Optional[List[str]] = None,
        debug_level: int = 0,
        debug_dir: Optional[Path] = None,
        sleep_level: int = 0,
        dry_run: bool = False,
        max_clip_count: int = 0,
        max_clip_errors: int = CLIP_ERROR_MAX,
    ) -> None:
        self.send_email = send_email
        self.sendmail = sendmail or ["/usr/sbin/sendmail"]
        self.debug_level = debug_level
        self.debug_dir = debug_dir
        self.sleep_level = sleep_level
        self.dry_run = dry_run
        self.max_clip_count = max_clip_count
        self.max_clip_errors = max_clip_errors

    def clip_for_account(self, account: Account) -> None:
        print(f"Clipping coupons for Safeway account {account.username}")
        try:
            swy = SafewayClient(account, self.debug_dir)
            clipped_offers: List[Offer] = []
            clip_errors: List[ClipError] = []
            offers = swy.get_offers()
            unclipped_offers = [
                o for o in offers if o.status == OfferStatus.Unclipped
            ]
            if not unclipped_offers:
                print("Nothing to do")
                return
            rjust_size = len(str(len(unclipped_offers)))
            for i, offer in enumerate(
                yield_delay(
                    unclipped_offers, self.sleep_level, self.debug_level
                )
            ):
                progress_count = (
                    f"({str(i + 1).rjust(rjust_size, ' ')}"
                    f"/{len(unclipped_offers)}) "
                )
                try:
                    if not self.dry_run:
                        swy.clip(offer)
                    print(f"{progress_count} Clipped {offer}")
                    clipped_offers.append(offer)
                    if (
                        self.max_clip_count
                        and len(clipped_offers) >= self.max_clip_count
                    ):
                        print(
                            "Clip maximum count of "
                            f"{self.max_clip_count} reached"
                        )
                        break
                except ClipError as e:
                    print(f"{progress_count} {e}")
                    clip_errors.append(e)
                    if (
                        self.max_clip_errors
                        and len(clip_errors) >= self.max_clip_errors
                    ):
                        raise TooManyClipErrors(
                            e,
                            clipped_offers=clipped_offers,
                            errors=clip_errors,
                        )

            print(f"Clipped {len(clipped_offers)} coupons")
            email_clip_results(
                self.sendmail,
                account,
                clipped_offers,
                error=None,
                clip_errors=clip_errors,
                debug_level=self.debug_level,
                send_email=self.send_email and not self.dry_run,
            )
        except Error as e:
            email_error(
                self.sendmail,
                account,
                error=e,
                debug_level=self.debug_level,
                send_email=self.send_email and not self.dry_run,
            )
            raise



import collections
import mimetypes
import os
import subprocess
from email.message import EmailMessage
from pathlib import Path
from typing import List, Optional

from .accounts import Account
from .errors import ClipError, Error, TooManyClipErrors
from .models import Offer


def _send_email(
    sendmail: List[str],
    account: Account,
    subject: str,
    mail_message: List[str],
    debug_level: int,
    send_email: bool,
    attachments: Optional[List[Path]] = None,
) -> None:
    mail_message_str = os.linesep.join(mail_message)
    if debug_level >= 1:
        if send_email:
            print(f"Sending email to {account.mail_to}")
        else:
            print(f"Would send email to {account.mail_to}")
        print(">>>>>>")
        print(mail_message_str)
        print("<<<<<<")
    if not send_email:
        return
    msg = EmailMessage()
    msg["To"] = account.mail_to
    msg["From"] = account.mail_from
    if subject:
        msg["Subject"] = subject
    msg.set_content(mail_message_str)
    for attachment in attachments or []:
        mt = mimetypes.guess_type(attachment.name)[0]
        main, sub = mt.split("/", 1) if mt else ("application", "octet-stream")
        msg.add_attachment(
            attachment.read_bytes(),
            filename=attachment.name,
            maintype=main,
            subtype=sub,
        )
    p = subprocess.Popen(
        sendmail + ["-f", account.mail_to, "-t"], stdin=subprocess.PIPE
    )
    p.communicate(bytes(msg.as_string(), "UTF-8"))


def email_clip_results(
    sendmail: List[str],
    account: Account,
    offers: List[Offer],
    error: Optional[Error],
    clip_errors: Optional[List[ClipError]],
    debug_level: int,
    send_email: bool,
) -> None:
    offers_by_type = collections.defaultdict(list)
    for offer in offers:
        offers_by_type[offer.offer_pgm].append(offer)
    mail_subject = f"Safeway coupons: {len(offers)} clipped"
    mail_message: List[str] = [
        f"Safeway account: {account.username}",
        f"Clipped {len(offers)} total:",
    ]
    for offer_type, offers_this_type in offers_by_type.items():
        mail_message.append(
            f"    {offer_type.name}: {len(offers_this_type)} coupons"
        )
    _send_email(
        sendmail, account, mail_subject, mail_message, debug_level, send_email
    )


def email_error(
    sendmail: List[str],
    account: Account,
    error: Error,
    debug_level: int,
    send_email: bool,
) -> None:
    mail_subject = f"Safeway coupons: {error.__class__.__name__} error"
    mail_message: List[str] = [
        f"Safeway account: {account.username}",
        f"Error: {error}",
    ]
    if isinstance(error, TooManyClipErrors) and error.clipped_offers:
        mail_message += ["Clipped coupons:", ""]
        for offer in error.clipped_offers:
            mail_message += str(offer)
    _send_email(
        sendmail,
        account,
        mail_subject,
        mail_message,
        debug_level,
        send_email,
        attachments=getattr(error, "attachments", None),
    )



import random
import time
from typing import Generator, Iterable, TypeVar

T = TypeVar("T")


def yield_delay(
    iterable: Iterable[T], sleep_level: int, debug_level: int
) -> Generator[T, None, None]:
    for count, item in enumerate(iterable):
        delay_time = 0.0
        if sleep_level < 2:
            if sleep_level < 1 and count and count % 12 == 0:
                if count % 48 == 0:
                    delay_time = random.uniform(15.0, 25.0)
                else:
                    delay_time = random.uniform(4.0, 8.0)
                delay_time = round(delay_time, 2)
            else:
                delay_time = random.uniform(0.3, 0.8)
        if delay_time:
            if debug_level >= 1 and delay_time >= 1:
                print(f"Waiting {delay_time} seconds")
            time.sleep(delay_time)
        yield item



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import List, Optional

import dataclasses_json


def datetime_encode(dt: datetime) -> str:
    return str(int(datetime.timestamp(dt)) * 1000)


def datetime_decode(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    return datetime.fromtimestamp(int(value) / 1000, timezone.utc)


class Model(dataclasses_json.DataClassJsonMixin):
    dataclass_json_config = dataclasses_json.config(
        letter_case=dataclasses_json.LetterCase.CAMEL,  # type: ignore
        undefined=dataclasses_json.Undefined.EXCLUDE,
        exclude=lambda f: f is None,  # type: ignore
    )["dataclasses_json"]


@dataclass
class OfferList(Model):
    offers: List[Offer] = field(
        metadata=dataclasses_json.config(
            field_name="companionGalleryOfferList"
        )
    )


class OfferStatus(Enum):
    Clipped = "C"
    Unclipped = "U"
    Unknown = "?"

    @classmethod
    def _missing_(cls, value: object) -> OfferStatus:
        return cls(cls.Unknown)


class OfferType(Enum):
    StoreCoupon = "SC"
    ManufacturerCoupon = "MF"
    PersonalizedDeal = "PD"
    Unknown = "?"

    @classmethod
    def _missing_(cls, value: object) -> OfferType:
        return cls(cls.Unknown)


@dataclass
class Offer(Model):
    offer_id: str
    status: OfferStatus
    name: str
    description: str
    start_date: datetime = field(
        metadata=dataclasses_json.config(
            encoder=datetime_encode, decoder=datetime_decode
        )
    )
    end_date: datetime = field(
        metadata=dataclasses_json.config(
            encoder=datetime_encode, decoder=datetime_decode
        )
    )
    offer_price: str
    offer_pgm: OfferType
    category_type: str
    image: str
    category: Optional[str] = None

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__} "
            f"{self.offer_pgm.value} {self.offer_id}: "
            f"[{self.offer_price}] {self.name}"
        )



import configparser
import itertools
import os
from typing import List, Optional

from .accounts import Account


class Config:
    @classmethod
    def load_accounts(cls, config_file: Optional[str] = None) -> List[Account]:
        account = cls.load_account_from_env()
        if account:
            return [account]
        if config_file:
            accounts = cls.load_accounts_from_config(config_file)
            if accounts:
                return accounts
        return []

    @classmethod
    def load_account_from_env(cls) -> Optional[Account]:
        username = os.environ.get("SAFEWAY_ACCOUNT_USERNAME")
        password = os.environ.get("SAFEWAY_ACCOUNT_PASSWORD")
        mail_to = os.environ.get("SAFEWAY_ACCOUNT_MAIL_TO")
        mail_from = os.environ.get("SAFEWAY_ACCOUNT_MAIL_TO")
        if username and password:
            return Account(
                username=username,
                password=password,
                mail_to=mail_to or username,
                mail_from=mail_from or username,
            )
        return None

    @classmethod
    def load_accounts_from_config(cls, config_file: str) -> List[Account]:
        config = configparser.ConfigParser()
        with open(config_file) as f:
            config.read_file(itertools.chain(["[_no_section]"], f))
        accounts: List[Account] = []
        mail_from = None
        for section in config.sections():
            if section in ["_no_section", "_global"]:
                if config.has_option(section, "email_sender"):
                    mail_from = config.get(section, "email_sender")
                continue
            mail_to = (
                config.get(section, "notify")
                if config.has_option(section, "notify")
                else None
            )
            username = str(section)
            accounts.append(
                Account(
                    username=username,
                    password=config.get(section, "password"),
                    mail_to=mail_to or username,
                    mail_from=mail_from or username,
                )
            )
        return accounts



from __future__ import annotations

from dataclasses import dataclass
from typing import List

from .models import Model, Offer, OfferType


@dataclass
class ClipRequestItem(Model):
    clip_type: str
    item_id: str
    item_type: OfferType


@dataclass
class ClipRequest(Model):
    items: List[ClipRequestItem]

    @classmethod
    def from_offer(cls, offer: Offer) -> ClipRequest:
        return cls(
            items=[
                ClipRequestItem(
                    clip_type=clip_type,
                    item_id=offer.offer_id,
                    item_type=offer.offer_pgm,
                )
                for clip_type in ["C", "L"]
            ]
        )


@dataclass
class ClipResponseItem(Model):
    clip_type: str
    item_id: str
    item_type: OfferType
    status: int
    clip_id: str
    checked: bool


@dataclass
class ClipResponse(Model):
    items: List[ClipResponseItem]

    @property
    def success(self) -> bool:
        return bool(self.items) and all(i.status == 1 for i in self.items)



import json
import random
from pathlib import Path
from typing import List, Optional

import requests

from .accounts import Account
from .errors import ClipError, HTTPError
from .methods import ClipRequest, ClipResponse
from .models import Offer, OfferList
from .session import BaseSession, LoginSession


class SafewayClient(BaseSession):
    def __init__(self, account: Account, debug_dir: Optional[Path]) -> None:
        self.session = LoginSession(account, debug_dir)
        self.requests.headers.update(
            {
                "Authorization": f"Bearer {self.session.access_token}",
                "X-SW" "Y_AP" "I_K" "EY": "em" "j" "ou",
                "X-SW" "Y_VERSION": "1.1",
                "X-SW" "Y-APPLICATION-TYPE": "web",
            }
        )

    def get_offers(self) -> List[Offer]:
        try:
            response = self.requests.get(
                "https://www.safeway.com/abs/pub/xapi"
                "/offers/companiongalleryoffer"
                f"?storeId={self.session.store_id}"
                f"&rand={random.randrange(100000,999999)}",
            )
            response.raise_for_status()
            return OfferList.from_dict(response.json()).offers
        except requests.exceptions.HTTPError as e:
            raise HTTPError(e, response) from e

    def clip(self, offer: Offer) -> None:
        request = ClipRequest.from_offer(offer)
        response: Optional[requests.Response] = None
        try:
            response = self.requests.post(
                "https://www.safeway.com/abs/pub/web/j4u/api/offers/clip"
                f"?storeId={self.session.store_id}",
                data=json.dumps(request.to_dict(encode_json=True)),
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()
            clip_response = ClipResponse.from_dict(response.json())
            if not clip_response.success:
                raise Exception(
                    f"Unsuccessful clip response for coupon {offer}"
                )
        except Exception as e:
            raise ClipError(e, response, offer) from e



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['safeway_coupons']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-json',
 'requests',
 'selenium>=4.10,<5.0',
 'undetected-chromedriver>=3.5.0,<4.0.0',
 'webdriver-manager>=3.8.6,<4.0.0']

entry_points = \
{'console_scripts': ['safeway-coupons = safeway_coupons.app:main']}

setup_kwargs = {
    'name': 'safeway-coupons',
    'version': '0.2.4',
    'description': 'Automatic coupon clipper for Safeway\'s online "Safeway for U" coupons',
    'long_description': '# Automatic Safeway coupon clipper\n\n[![PyPI](https://img.shields.io/pypi/v/safeway-coupons)][pypi]\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/safeway-coupons)][pypi]\n[![Build](https://img.shields.io/github/checks-status/smkent/safeway-coupons/main?label=build)][gh-actions]\n[![codecov](https://codecov.io/gh/smkent/safeway-coupons/branch/main/graph/badge.svg)][codecov]\n[![GitHub stars](https://img.shields.io/github/stars/smkent/safeway-coupons?style=social)][repo]\n\n**safeway-coupons** is a script that will log in to an account on safeway.com,\nand attempt to select all of the "Safeway for U" electronic coupons on the site\nso they don\'t have to each be clicked manually.\n\n## Design notes\n\nSafeway\'s sign in page is protected by a web application firewall (WAF).\nsafeway-coupons performs authentication using a headless instance of Google\nChrome. Authentication may fail based on your IP\'s reputation, either by\npresenting a CAPTCHA or denying sign in attempts altogether. safeway-coupons\ncurrently does not have support for prompting the user to solve CAPTCHAs.\n\nOnce a signed in session is established, coupon clipping is performed using HTTP\nrequests via [requests][requests].\n\n## Installation and usage with Docker\n\nA Docker container is provided which runs safeway-coupons with cron. The cron\nschedule and your Safeway account details may be configured using environment\nvariables, or with an accounts file.\n\nExample `docker-compose.yaml` with configuration via environment variables:\n\n```yaml\nversion: "3.7"\n\nservices:\n  safeway-coupons:\n    image: ghcr.io/smkent/safeway-coupons:latest\n    environment:\n      CRON_SCHEDULE: "0 2 * * *"  # Run at 2:00 AM UTC each day\n      # TZ: Antarctica/McMurdo  # Optional time zone to use instead of UTC\n      SMTPHOST: your.smtp.host\n      SAFEWAY_ACCOUNT_USERNAME: your.safeway.account.email@example.com\n      SAFEWAY_ACCOUNT_PASSWORD: very_secret\n      SAFEWAY_ACCOUNT_MAIL_FROM: your.email@example.com\n      SAFEWAY_ACCOUNT_MAIL_TO: your.email@example.com\n      # EXTRA_ARGS: --debug  # Optional\n    restart: unless-stopped\n```\n\nExample `docker-compose.yaml` with configuration via accounts file:\n\n```yaml\nversion: "3.7"\n\nservices:\n  safeway-coupons:\n    image: ghcr.io/smkent/safeway-coupons:latest\n    environment:\n      CRON_SCHEDULE: "0 2 * * *"  # Run at 2:00 AM UTC each day\n      # TZ: Antarctica/McMurdo  # Optional time zone to use instead of UTC\n      SMTPHOST: your.smtp.host\n      SAFEWAY_ACCOUNTS_FILE: /accounts_file\n      # EXTRA_ARGS: --debug  # Optional\n    restart: unless-stopped\n    volumes:\n      - path/to/safeway_accounts_file:/accounts_file:ro\n```\n\nStart the container by running:\n\n```console\ndocker-compose up -d\n```\n\nDebugging information can be viewed in the container log:\n\n```console\ndocker-compose logs -f\n```\n\n## Installation from PyPI\n\n### Prerequisites\n\n* Google Chrome (for authentication performed via Selenium).\n* Optional: `sendmail` (for email support)\n\n### Installation\n\n[safeway-coupons is available on PyPI][pypi]:\n\n```console\npip install safeway-coupons\n```\n\n### Usage\n\nFor best results, run this program once a day or so with a cron daemon.\n\nFor full usage options, run\n\n```console\nsafeway-coupons --help\n```\n\n### Configuration\n\n**safeway-coupons** can clip coupons for one or more Safeway accounts in a\nsingle run, depending on the configuration method used.\n\nIf a sender email address is configured, a summary email will be sent for each\nSafeway account via `sendmail`. The email recipient defaults to the Safeway\naccount email address, but can be overridden for each account.\n\nAccounts are searched via the following methods in the listed order. Only one\naccount configuration method may be used at a time.\n\n#### With environment variables\n\nA single Safeway account can be configured with environment variables:\n\n* `SAFEWAY_ACCOUNT_USERNAME`: Account email address (required)\n* `SAFEWAY_ACCOUNT_PASSWORD`: Account password (required)\n* `SAFEWAY_ACCOUNT_MAIL_FROM`: Sender address for email summary\n* `SAFEWAY_ACCOUNT_MAIL_TO`: Recipient address for email summary\n\n#### With config file\n\nMultiple Safeway accounts can be provided in an ini-style config file, with a\nsection for each account. For example:\n\n```ini\nemail_sender = sender@example.com   ; optional\n\n[safeway.account@example.com]       ; required\npassword = 12345                    ; required\nnotify = your.email@example.com     ; optional\n```\n\nProvide the path to your config file using the `-c` or `--accounts-config`\noption:\n\n```console\nsafeway-coupons -c path/to/config/file\n```\n\n## Development\n\n### [Poetry][poetry] installation\n\nVia [`pipx`][pipx]:\n\n```console\npip install pipx\npipx install poetry\npipx inject poetry poetry-dynamic-versioning poetry-pre-commit-plugin\n```\n\nVia `pip`:\n\n```console\npip install poetry\npoetry self add poetry-dynamic-versioning poetry-pre-commit-plugin\n```\n\n### Development tasks\n\n* Setup: `poetry install`\n* Run static checks: `poetry run poe lint` or\n  `poetry run pre-commit run --all-files`\n* Run static checks and tests: `poetry run poe test`\n\n---\n\nCreated from [smkent/cookie-python][cookie-python] using\n[cookiecutter][cookiecutter]\n\n[codecov]: https://codecov.io/gh/smkent/safeway-coupons\n[cookie-python]: https://github.com/smkent/cookie-python\n[cookiecutter]: https://github.com/cookiecutter/cookiecutter\n[gh-actions]: https://github.com/smkent/safeway-coupons/actions?query=branch%3Amain\n[pipx]: https://pypa.github.io/pipx/\n[poetry]: https://python-poetry.org/docs/#installation\n[pypi]: https://pypi.org/project/safeway-coupons/\n[repo]: https://github.com/smkent/safeway-coupons\n[requests]: https://requests.readthedocs.io/en/latest/\n',
    'author': 'Stephen Kent',
    'author_email': 'smkent@smkent.net',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/smkent/safeway-coupons',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import requests

from .accounts import Account
from .models import Offer


@dataclass
class Error(Exception):
    exception: Exception


@dataclass
class AuthenticationFailure(Error):
    account: Account
    attachments: Optional[List[Path]] = None

    def __str__(self) -> str:
        return f"Authentication Failure ({self.exception})"


@dataclass
class HTTPError(Error):
    response: requests.Response

    def __str__(self) -> str:
        return str(self.exception)


@dataclass
class ClipError(Error):
    response: Optional[requests.Response]
    offer: Offer

    def __str__(self) -> str:
        return f"Clip Error for {self.offer} ({self.exception})"


@dataclass
class TooManyClipErrors(Error):
    clipped_offers: List[Offer]
    errors: List[ClipError]

    def __str__(self) -> str:
        return (
            f"Exited after {len(self.errors)} clip errors "
            f"(Last: {self.exception})"
        )



import argparse
import shlex
import sys
import traceback
from http.client import HTTPConnection
from pathlib import Path

from .config import Config
from .errors import Error
from .safeway import SafewayCoupons


def _parse_args() -> argparse.Namespace:
    description = 'Automatic coupon clipper for "Safeway for U" coupons'
    arg_parser = argparse.ArgumentParser(description=description)
    arg_parser.add_argument(
        "-c",
        "--accounts-config",
        dest="accounts_config",
        metavar="file",
        help=(
            "Path to configuration file containing Safeway "
            "accounts information"
        ),
    )
    arg_parser.add_argument(
        "-D",
        "--debug-dir",
        dest="debug_dir",
        metavar="directory",
        default=".",
        help=(
            "Destination directory for debug output files, "
            "such as browser screenshots (default: %(default)s)"
        ),
    )
    arg_parser.add_argument(
        "-d",
        "--debug",
        dest="debug_level",
        action="count",
        default=0,
        help=(
            "Print debugging information on stdout. Specify "
            "twice to increase verbosity."
        ),
    )
    arg_parser.add_argument(
        "-m",
        "--max-clip",
        dest="max_clip_count",
        type=int,
        default=0,
        metavar="number",
        help="Maximum number of coupons to clip (default: all)",
    )
    arg_parser.add_argument(
        "--sendmail",
        metavar="path/to/sendmail",
        dest="sendmail",
        type=shlex.split,
        default="/usr/sbin/sendmail",
        help=(
            "Path to sendmail and any additional arguments to use when "
            "sending email (default: %(default)s)"
        ),
    )
    arg_parser.add_argument(
        "-n",
        "--no-email",
        dest="send_email",
        action="store_false",
        help="Don't send results email",
    )
    arg_parser.add_argument(
        "-p",
        "--pretend",
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="Print coupons that would be clipped, but don't clip them",
    )
    arg_parser.add_argument(
        "-S",
        "--no-sleep",
        "--no-delay",
        dest="sleep_level",
        action="count",
        default=0,
        help=(
            "Don't wait between long requests. Specify twice to never wait."
        ),
    )
    arg_parser.add_argument(
        "-E",
        "--continue-on-error",
        dest="continue_on_error",
        action="store_true",
        help="Continue clipping coupons for the next account on error",
    )
    return arg_parser.parse_args()


def main() -> None:
    args = _parse_args()
    accounts = Config.load_accounts(config_file=args.accounts_config)
    if not accounts:
        print("Error: No Safeway account(s) configured", file=sys.stderr)
        sys.exit(1)
    if args.debug_level >= 2:
        HTTPConnection.debuglevel = 1
    sc = SafewayCoupons(
        send_email=args.send_email,
        sendmail=args.sendmail,
        debug_level=args.debug_level,
        debug_dir=Path(args.debug_dir) if args.debug_dir else None,
        sleep_level=args.sleep_level,
        dry_run=args.dry_run,
        max_clip_count=args.max_clip_count,
    )
    errors = 0
    try:
        for account in accounts:
            try:
                sc.clip_for_account(account)
            except Error:
                if not args.continue_on_error:
                    raise
                traceback.print_exc()
                errors += 1
        if errors:
            print(f"Error clipping coupons for {errors} account(s)")
            sys.exit(1)
    except Exception as e:
        if args.debug_level:
            raise
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)
